/*

  File Chebyshev_polynomials.ccp:
  solves the equation:

     (1-x^2) * d^2 Y / dx^2 - x * d Y/dx + n^2 * Y = 0

  as a boundary value problem, with finite differences centred schemes.

*/


#include <iostream>
#include <fstream>
#include <cmath>
#include "LUtrid.h"
#include "LUvect.h"
using namespace std;


/*

  The following global variables identify the characteristics of the
  problem, depending on the boundary conditions, like the effective
  dimension of the problem "dim" (dimension of the matrix to be inverted),
  the index of the initial and end points on the grid (j_init, j_end)
  as well as the quantities:
  Ya (value of the unknown on the left boundary),
  Yb (value of the unknown on the right boundary),
  Ypa (value of the first derivative of the unknown on the left boundary),
  Ypb (value of the first derivative of the unknown on the right boundary).

  The variable bct identifies the combination of known boundary conditions:
     bct = 1 => Y(a) and Y(b) are known;
     bct = 2 => Y'(a) and Y'(b) are known.
     bct = 3 => Y'(a) and Y(b) are known;
     bct = 4 => Y(a) and Y'(b) are known;

*/

int N;                     // Number of intervals
int bct;                   // Type of boundary conditions (see above!)
int dim;                   // Effective dimension of the problem
int j_init, j_end;         // Effective values of j for solving the tridiagonal system 
double Ya, Yb, Ypa, Ypb;   // Numerical value of the boundary conditions for Y(a), Y(b),
                           // Y'(a) and Y'(b)
int nn;


// Functions defining the parameters of the equation

double d( double x )
{
  return( 1.0 - x * x );
}


double f( double x )
{
  return( - x );
}


double g( double x )
{
  return( nn * nn );
}


double h( double x )
{
  return( 0.0 );
}


// Main program

int main()
{
  cout << "Solution of the differential equation:" << endl;
  cout << "(1-x^2) * d^2 Y / dx^2 - x * d Y / dx + n^2 * Y = 0" << endl;


  // Input parameters

  cout << "Number of intervals to use (the number of grid-points will be N+1!): ";
  cin >> N;

  cout << "Type of boundary conditions (bct = 1, 2, 3, 4): ";
  cin >> bct;

  cout << "n: ";
  cin >> nn;


  // Selects the variables according with the type of imposed boundary conditions
  // Note that here the points x_0, ..., x_N correspond to elements 1, ..., N+1!
  
  switch ( bct )
    {

    case 1:      // Y(a) and Y(b) known:
      j_init = 2;
      j_end = N;
      Ya  = pow( -1.0, nn );
      Yb  = 1.0;
      Ypa = 0.0;
      Ypb = 0.0;
      break;

    case 2:      // Y'(a) and Y'(b) known
      j_init = 1;
      j_end = N + 1;
      Ya  = 0.0;
      Yb  = 0.0;
      Ypa = pow( -1.0, nn ) * nn * nn;
      Ypb = nn * nn;
      break;

    case 3:      // Y'(a) and Y(b) known
      j_init = 1;
      j_end = N;
      Ya  = 0.0;
      Yb  = 1.0;
      Ypa = pow( -1.0, nn ) * nn * nn;
      Ypb = 0.0;
      break;

    case 4:      // Y(a) and Y'(b) known
      j_init = 2;
      j_end = N + 1;
      Ya  = pow( -1.0, nn );
      Yb  = 0.0;
      Ypa = 0.0;
      Ypb = nn * nn;
      break;

    }


  // Dimension of the tridiagonal matrix (according to the boundary conditions!)

  dim = j_end - j_init + 1;


  // Defines the grid-points x_j

  LUvect x = LUvect( N + 1 );
  double a, b;
  a = -1.0;     // Set the value for the left boundary a
  b = +1.0 ;    // Set the value for the right boundary b
  double dx = ( b - a ) / N;     // Note: the number of intervals is N-1!
  double dx2 = dx * dx;
  double ddx = 2.0 * dx;
  for ( int j = 1; j <= N + 1; j++ ) x.setElement( j, a + ( j - 1 ) * dx );


  // Defines the tridiagonal matrix and the RHS

  LUtrid D = LUtrid( dim );
  LUvect sm = LUvect( dim );


  // What appears in the first and last rows of the matrix and on the RHS
  // depends strictly on the boundary conditions

  double xj_init = x.getElement( j_init );
  double xj_end  = x.getElement( j_end );
  double la_j, mu_j, xi_j;
  
  switch ( bct )
    {
    case 1:
      la_j = d( xj_init ) / dx2 - f( xj_init ) / ddx;
      mu_j = - 2.0 * d( xj_init ) / dx2 + g( xj_init );
      xi_j = d( xj_init ) / dx2 + f( xj_init ) / ddx;
      D.setElement( 1, 1, mu_j );
      D.setElement( 1, 2, xi_j );
      sm.setElement( 1, h( xj_init ) - Ya * la_j );

      la_j = d( xj_end ) / dx2 - f( xj_end ) / ddx;
      mu_j = - 2.0 * d( xj_end ) / dx2 + g( xj_end );
      xi_j = d( xj_end ) / dx2 + f( xj_end ) / ddx;
      D.setElement( dim, dim - 1, la_j );
      D.setElement( dim, dim, mu_j );
      sm.setElement( dim, h( xj_end ) - Yb * xi_j );
      break;

    case 2:
      la_j = d( xj_init ) / dx2 - f( xj_init ) / ddx;
      mu_j = - 2.0 * d( xj_init ) / dx2 + g( xj_init );
      xi_j = d( xj_init ) / dx2 + f( xj_init ) / ddx;
      D.setElement( 1, 1, mu_j );
      D.setElement( 1, 2, ( xi_j + la_j ) );
      sm.setElement( 1, h( xj_init ) - ddx * Ypa * la_j );

      la_j = d( xj_end ) / dx2 - f( xj_end ) / ddx;
      mu_j = - 2.0 * d( xj_end ) / dx2 + g( xj_end );
      xi_j = d( xj_end ) / dx2 + f( xj_end ) / ddx;
      D.setElement( dim, dim - 1, ( xi_j + la_j ) );
      D.setElement( dim, dim, mu_j );
      sm.setElement( dim, h( xj_end ) - ddx * Ypb * xi_j );
      break;

    case 3:
      la_j = d( xj_init ) / dx2 - f( xj_init ) / ddx;
      mu_j = - 2.0 * d( xj_init ) / dx2 + g( xj_init );
      xi_j = d( xj_init ) / dx2 + f( xj_init ) / ddx;
      D.setElement( 1, 1, mu_j );
      D.setElement( 1, 2, ( xi_j + la_j ) );
      sm.setElement( 1, h( xj_init ) - ddx * Ypa * la_j );

      la_j = d( xj_end ) / dx2 - f( xj_end ) / ddx;
      mu_j = - 2.0 * d( xj_end ) / dx2 + g( xj_end );
      xi_j = d( xj_end ) / dx2 + f( xj_end ) / ddx;
      D.setElement( dim, dim - 1, la_j );
      D.setElement( dim, dim, mu_j );
      sm.setElement( dim, h( xj_end ) - Yb * xi_j );
      break;

    case 4:
      la_j = d( xj_init ) / dx2 - f( xj_init ) / ddx;
      mu_j = - 2.0 * d( xj_init ) / dx2 + g( xj_init );
      xi_j = d( xj_init ) / dx2 + f( xj_init ) / ddx;
      D.setElement( 1, 1, mu_j );
      D.setElement( 1, 2, xi_j );
      sm.setElement( 1, h( xj_init ) - Ya * la_j );

      la_j = d( xj_end ) / dx2 - f( xj_end ) / ddx;
      mu_j = - 2.0 * d( xj_end ) / dx2 + g( xj_end );
      xi_j = d( xj_end ) / dx2 + f( xj_end ) / ddx;
      D.setElement( dim, dim - 1, ( xi_j + la_j ) );
      D.setElement( dim, dim, mu_j );
      sm.setElement( dim, h( xj_end ) - ddx * Ypb * xi_j );
      break;
    }


  // Definition of the matrix for the central rows that do not depend on the
  // chosen boundary conditions

  int jj = 2;
  for ( int j = j_init + 1; j <= j_end - 1; j++ )
    {
      double xj = x.getElement( j );
      la_j = d( xj ) / dx2 - f( xj ) / ddx;
      mu_j = - 2.0 * d( xj ) / dx2 + g( xj );
      xi_j = d( xj ) / dx2 + f( xj ) / ddx;
      D.setElement( jj, jj - 1, la_j );
      D.setElement( jj, jj, mu_j );
      D.setElement( jj, jj + 1, xi_j );
      sm.setElement( jj, h( xj ) );
      jj++;
    }


  // LU factorization of the matrix

  D.luFact();


  // Solves the LU system

  LUvect Y = LUvect( dim );
  D.luSolve( Y, sm );


  // Prints the result

  // Read the file name as input

  char fname[250];
  cout << "File name for the results: ";
  cin >> fname;

  // Opens the output file and sets the maximum precision for writing

  ofstream outFile;
  outFile.open( fname );
  outFile.precision( 15 );

  // If Y(a) is known, write it to the file

  if ( ( bct == 1 ) || ( bct == 4 ) ) outFile << x.getElement( 1 ) << "\t\t" << Ya << endl;

  // Writes the computed solution to the file

  jj = 1;
  for ( int j = j_init; j <= j_end; j++ )
    outFile << x.getElement( j ) << "\t\t" << Y.getElement( jj++ ) << endl;

  // If Y(b) is known, write it to the file

  if ( ( bct == 1 ) || ( bct == 3 ) ) outFile << x.getElement( N + 1 ) << "\t\t" << Yb << endl;

  // Closes the output file and ends the program

  outFile.close();
  return( 0 );
}
