/*

  File LUtrid.cpp:

  contains the definitions for the member functions of the class
  LUtrid that solve a tridiagonal system 3xn, carries out its LU
  factorization and solves the system.

*/


#include <iostream>
#include "LUtrid.h"
#include <stdlib.h>
using namespace std;


/*

 LUtrid(): class constructor

*/

LUtrid::LUtrid( int N )
{
  dim = N;
  mat = new double * [3];
  for ( int i=0; i < 3; i++ ) mat[i] = new double[N];
  tmp = new double[N];           // Temporary vector to store the RHS of the Ux=y system.
}


/*

  luFact( ): computes the LU factorization of a matrix a[3][dim]
  according to Thomas' algorithm.

*/


void LUtrid::luFact( )
{

  int i;

  for ( i = 1; i < dim; i++ )
    {
      mat[0][i] = mat[0][i] / mat[1][i-1];
      mat[1][i] = mat[1][i] - mat[0][i] * mat[2][i-1];
    }

}


/*

  luSolve( x, b ): solves a generic tridiagonal system A*x=b
  according to Thomas' algorithm.

*/


void LUtrid::luSolve( LUvect x, LUvect b )
{

  int i;

  // Forward substitution

  tmp[0] = b.getElement( 1 );    // First element
  for ( i = 2; i <= dim; i++ )
    {
      tmp[i-1] = b.getElement( i ) - mat[0][i-1] * tmp[i-2];
    }

  // Backward substitution

  x.setElement( dim, tmp[dim-1] / mat[1][dim-1] );
  for ( i = dim - 1; i > 0; i-- )
    {
      x.setElement( i, ( tmp[i-1] - mat[2][i-1] * x.getElement( i+1 ) ) / mat[1][i-1] );
    }

}


/*

  Function setElement:
  sets the element a_ij of the matrix to the given value.
  Note that, in this case, the position i-j of the element
  refers to the original matrix, not the one ordered by diagonals!

*/

void LUtrid::setElement( int i, int j, double a_ij )
{
  if ( ( ( i > 0 ) && ( i <= dim ) ) && ( ( j > 0 ) && ( j <= dim ) ) )
    {
      if ( i-1 == j ) mat[0][i-1] = a_ij;
      if ( i == j )   mat[1][i-1] = a_ij;
      if ( i+1 == j ) mat[2][i-1] = a_ij;
    }

  // Ensures that b_1 = c_N = 0.0 anyway...
  mat[0][0] = 0.0;
  mat[2][dim-1] = 0.0;
}


/*

  Function getElement:
  returns the value of the i-th, j-th element of the matrix.
  Note that, also in this case, the position i-j refers to the
  original matrix, not the one ordered by diagonals!

*/

double LUtrid::getElement( int i, int j )
{
  double elem;

  if ( ( ( i > 0 ) && ( i <= dim ) ) && ( ( j > 0 ) && ( j <= dim ) ) )
    {
      elem = 0.0;
      if ( i-1 == j ) elem = mat[0][i-1];
      if ( i == j )   elem = mat[1][i-1];
      if ( i+1 == j ) elem = mat[2][i-1];
    }
  return( elem );
}


/*

  Function matVectProduct:
  computes the row by column product of the current  tridiagonal matrix 
  times a vector. The vector to multiply is the first argument, while
  the result is stored in the second argument.

*/

void LUtrid::matVectProduct( LUvect u, LUvect v )
{
  double sum;

  // Computes the product for the first row: v_1 = a_1 * u_1 + c_1 * u_2

  sum = mat[1][0] * u.getElement( 1 ) + mat[2][0] * u.getElement( 2 );
  v.setElement( 1, sum );

  // Computes the central rows

  for ( int i = 2; i <= dim - 1; i++ )
    {
      sum = mat[0][i-1] * u.getElement( i - 1 ) + mat[1][i-1] * u.getElement( i ) + mat[2][i-1] * u.getElement( i + 1 );
      v.setElement( i, sum );
    }

  // Computes the final row

  sum = mat[0][dim-1] * u.getElement( dim - 1 ) + mat[1][dim-1] * u.getElement( dim );
  v.setElement( dim, sum );

}


/*

  Function duplicateMat:
  copies a given tridiagonal matrix in the current matrix.

*/

void LUtrid::duplicateMat( LUtrid dmat )
{
  for ( int i = 0; i < dim; i++ )
    {
      this->mat[0][i] = dmat.mat[0][i];
      this->mat[1][i] = dmat.mat[1][i];
      this->mat[2][i] = dmat.mat[2][i];
    }

}
